@@ -1,14 +1,18 @@
 /*************************************************
 Documentation of symbols defined by Math::MPFR
 
-NV_IS_LONG_DOUBLE        : Automatically defined by Makefile.PL iff
+NV_IS_LONG_DOUBLE        : Automatically defined by Makefile.PL if
                            $Config{nvtype} is 'long double'.
 
-NV_IS_FLOAT128           : Automatically defined by Makefile.PL iff
+NV_IS_FLOAT128           : Automatically defined by Makefile.PL if
                            $Config{nvtype} is __float128
                            If NV_IS_FLOAT128 is defined we include the
                            quadmath.h header.
 
+NV_IS_53_BIT             : Defined only when $Config{nvtype} is 'double' or
+                           when $Config{nvtype} is a 'long double' that's
+                           identical to the double.
+
 MPFR_WANT_FLOAT128       : Defined by Makefile.PL if $have_float128 is
                            set to a true value. $have_float128 can be set
                            to a true value by either editing the Makefile.PL
@@ -35,29 +39,51 @@ CAN_PASS_FLOAT128        : Defined only when both MPFR_WANT_FLOAT128 and
                            4.0.0.)
                            DANGER: The assumption is that if MPFR_WANT_FLOAT128
                            is defined then the mpfr library has been built
-                           with __float128 support, which won't be the case if
-                           the mpfr library wasn't configured with
-                           '--enable-float128'.
-                           I haven't yet found a way of managing this - it's
-                           instead left up to the person building Math::MPFR to
-                           NOT define MATH_MPFR_WANT_FLOAT128 unless mpfr WAS
-                           configured with --enable-float128.
+                           with __float128 support, which may not be the case.
+                           Hopefully the configure probing done by the
+                           Makefile.PL will get it right.
 
 MPFR_WANT_DECIMAL_FLOATS : The symbol needs to be defined (before mpfr.h is
-                           included) in order to enable _Decimal64 support.
+                           included) in order to enable _Decimal64 and/or
+                           _Decimal128 support.
                            Hence we define it in the Makefile.PL by setting
-                           $have_decimal64 to a true value. $have_decimal64
-                           can also be set to a true value by specifying
-                           D64=1 in the Makefile.PL's @ARGV.
+                           $have_decimal64 or setting $have_decimal128 to a true
+                           value.
+                           $have_decimal64 can be forcibly set to a true value by
+                           specifying D64=1 in the Makefile.PL's @ARGV.
+                           And $have_decimal128 can likewise be set to a true value
+                           by specifying D128=1 in the Makefile.PL's @ARGV.
                            $have_decimal64 must not be set to a true value
                            if the mpfr library has not been built with
                            _Decimal64 support.
-                           We define the symbol solely to make use of the
-                           Math::Decimal64-Math::MPFR interface. Otherwise
-                           there's no point (apparent to me) in defining it.
+                           And $have_decimal128 must not be set to a true value
+                           if the mpfr library has not been built with
+                           _Decimal128 support.
+                           We define the symbol solely to enable a Math::MPFR
+                           interface with Math::Decimal64 and/or Math::Decimal128.
+                           Otherwise there's no point (apparent to me) in defining
+                           it.
+
+MPFR_WANT_DECIMAL64      : Defined by the Makefile.PL only if support for the
+                            _Decimal64 type in the mpfr library is detected.
+                           Can be forcibly defined by specifying D64=1 in the
+                           Makefile.PL's @ARGV.
+                           Can be forcibly not defined by specifying D64=0
+                           in the Makefile.PL's @ARGV.
+                           We define the symbol solely to enable a
+                           Math::Decimal64-Math::MPFR interface.
+
+MPFR_WANT_DECIMAL128     : Defined by the Makefile.PL only if support for the
+                           _Decimal128 type in the mpfr library is detected.
+                           Can be forcibly defined by specifying D128=1 in the
+                           Makefile.PL's @ARGV.
+                           Can be forcibly not defined by specifying D128=0
+                           in the Makefile.PL's @ARGV.
+                           We define the symbol solely to enable a
+                           Math::Decimal128-Math::MPFR interface.
 
 HAVE_IEEE_754_LONG_DOUBLE :Used only by the test suite.
-                           Defined by Makefile.PL iff
+                           Defined by Makefile.PL if
                            ($Config{longdblkind} == 1 ||
                             $Config{longdblkind} == 2)
                            This implies that long double is the quad (128-bit)
@@ -65,7 +91,7 @@ HAVE_IEEE_754_LONG_DOUBLE :Used only by the test suite.
 
 HAVE_EXTENDED_PRECISION_LONG_DOUBLE :
                            Used only by the test suite.
-                           Defined by Makefile.PL iff
+                           Defined by Makefile.PL if
                            ($Config{longdblkind} == 3 ||
                             $Config{longdblkind} == 4)
                            This implies that nvtype is the extended
@@ -78,14 +104,17 @@ REQUIRED_LDBL_MANT_DIG   : Defined to float.h's LDBL_MANT_DIG unless
                            This is needed to ensure that the mpfr value is
                            an accurate rendition of the double-double value.
 
-MAXIMUM_ALLOWABLE_BASE   : Defined to 62 iff mpfr version >= 3.0.0.
-                           Else defined to 32.
-
-CHECK_ROUNDING_VALUE     : Macro that checks (on mpfr-versions 2.x.x only)
+CHECK_ROUNDING_VALUE     : Macro that checks (on pre-4.0.0 mpfr versions only)
                            that the rounding value provided is in the
-                           allowable range of 0-3 inclusive.
-                           (The range has been extended for versions 3.0.0
-                           and later.)
+                           allowable range of 0-4 inclusive.
+                           (On 2.x.x versions the allowable range is only 0-3,
+                           but we don't support those versions anyway.)
+
+CHECK_INPUT_BASE         : Macro that checks that the base (where specified)
+                           is in the accepted range.
+
+CHECK_OUTPUT_BASE        : Macro that checks that the base (where specified)
+                           is in the accepted range.
 
 DEAL_WITH_NANFLAG_BUG    : Macro that corrects certain failures (in mpfr
                            versions prior to 3.1.4) to set the NaN flag.
@@ -117,23 +146,42 @@ IVSIZE_BITS              : Defined only if MATH_MPFR_NEED_LONG_LONG_INT is
                            Currently, I think this symbol will only ever be
                            either undefined or set to 64 - and I suspect
                            that it could (currently) be replaced with a hard
-                           code 64 wherever it occurs in the code.
+                           coded 64 wherever it occurs in the code.
 
 _WIN32_BIZARRE_INFNAN    : Defined (on Windows only) when the perl version
                            (as expressed by $]) is less than 5.022.
-                           These earlier perl versions generally stringified
-                           NaNs as (-)1.#IND and Infs as (-)1.#INF.
+                           These earlier perl versions had bizarre strings
+                           representing NaNs (eg 1.#IND) and Infs (eg 1.#INF)
+                           on Win32.
+
+LD_SUBNORMAL_BUG         : Defined for mpfr-3.1.4 and earlier if and only if
+                           LDBL_MANT_DIG == 64
+                           (The bug is in mpfr_get_ld)
+
+FALLBACK_NOTIFY          : If defined, $Math::MPFR::doubletoa_fallback
+                           (initially set to 0) will be incremented by 1 on
+                           those rare occasions where grisu3 fails and
+                           falls back to the fallback routine.
+                           For more details, see the doubletoa documentation.
 
 *************************************************/
 
 #include <stdio.h>
 
-#if defined MATH_MPFR_NEED_LONG_LONG_INT
+#if defined(MATH_MPFR_NEED_LONG_LONG_INT)
 #ifndef _MSC_VER
 #include <inttypes.h>
 #endif
 #endif
 
+/*
+ * In mpfr-4.1.0, the _Float128 type is exposed in mpfr.h if MPFR_WANT_FLOAT128 is defined.
+ * We fall back to defining it to __float128 if the _Float128 type is unknown.
+*/
+
+#if defined(MPFR_WANT_FLOAT128) && defined(__GNUC__) && !defined(__FLT128_MAX__) && !defined(_BITS_FLOATN_H)
+#define _Float128 __float128
+#endif
 
 #include <gmp.h>
 #include <mpfr.h>
@@ -153,27 +201,44 @@ typedef __float128 float128;
 #endif
 #endif
 
+#if defined(MPFR_WANT_DECIMAL128)
+#if defined(__MINGW64__) || (defined(DEBUGGING) && defined(NV_IS_DOUBLE))
+typedef _Decimal128 D128 __attribute__ ((aligned(8)));
+#else
+typedef _Decimal128 D128;
+#endif
+#endif
+
+
+#if (!defined(MPFR_VERSION) || MPFR_VERSION <= 196868) && LDBL_MANT_DIG == 64
+#define LD_SUBNORMAL_BUG 1
+#endif
+
 #if LDBL_MANT_DIG == 106
 #define REQUIRED_LDBL_MANT_DIG 2098
 #else
 #define REQUIRED_LDBL_MANT_DIG LDBL_MANT_DIG
 #endif
 
-#if defined(MPFR_VERSION_MAJOR) && MPFR_VERSION_MAJOR >= 3
-#define MAXIMUM_ALLOWABLE_BASE 62
+#if (!defined(NV_IS_FLOAT128) && !defined(NV_IS_LONG_DOUBLE)) || (defined(NV_IS_LONG_DOUBLE) && REQUIRED_LDBL_MANT_DIG == 53)
+#define NV_IS_53_BIT 1
+#endif
+
+#define CHECK_INPUT_BASE \
+     if(SvIV(base) < 0 || SvIV(base) > 62 || SvIV(base) == 1) {
+
+#if MPFR_VERSION >= 262400 /* Allowable range of base has been expanded */
+#define CHECK_OUTPUT_BASE \
+     if(SvIV(base) < -36 || SvIV(base) > 62 || abs(SvIV(base)) < 2 )  {
 #else
-#define MAXIMUM_ALLOWABLE_BASE 36
+#define CHECK_OUTPUT_BASE \
+     if(SvIV(base) < 2 || SvIV(base) > 62)                    {
 #endif
 
 /* Don't use CHECK_ROUNDING_VALUE macro with Rmpfr_set_NV      *
  * (as this function's "round" arg is "unsigned int", not SV*) */
 
-#if MPFR_VERSION_MAJOR < 3
-#define CHECK_ROUNDING_VALUE \
- if((mp_rnd_t)SvUV(round) > 3) \
-  croak("Illegal rounding value supplied for this version (%s) of the mpfr library", MPFR_VERSION_STRING);
-
-#elif MPFR_VERSION_MAJOR < 4
+#if MPFR_VERSION_MAJOR < 4
 #define CHECK_ROUNDING_VALUE \
  if((mp_rnd_t)SvUV(round) > 4) \
   croak("Illegal rounding value supplied for this version (%s) of the mpfr library", MPFR_VERSION_STRING);
@@ -197,6 +262,52 @@ typedef __float128 float128;
          if(SvIV(get_sv("Math::MPFR::NNW", 0))) \
            warn("string used in %s contains non-numeric characters"
 
+#define BITSEARCH_4 \
+          if(tmp & 8) {				\
+            subnormal_prec_adjustment += 1;	\
+            break;				\
+          }					\
+          if(tmp & 4) {				\
+            subnormal_prec_adjustment += 2;	\
+            break;				\
+          }					\
+          if(tmp & 2) {				\
+            subnormal_prec_adjustment += 3;	\
+            break;				\
+          }					\
+          subnormal_prec_adjustment += 4;
+
+
+#define BITSEARCH_8 \
+          if(tmp & 128) {			\
+            subnormal_prec_adjustment += 1;	\
+            break;				\
+          }					\
+          if(tmp & 64) {			\
+            subnormal_prec_adjustment += 2;	\
+            break;				\
+          }					\
+          if(tmp & 32) {			\
+            subnormal_prec_adjustment += 3;	\
+            break;				\
+          }					\
+          if(tmp & 16) {			\
+            subnormal_prec_adjustment += 4;	\
+            break;				\
+          }					\
+          if(tmp & 8) {				\
+            subnormal_prec_adjustment += 5;	\
+            break;				\
+          }					\
+          if(tmp & 4) {				\
+            subnormal_prec_adjustment += 6;	\
+            break;				\
+          }					\
+          if(tmp & 2) {				\
+            subnormal_prec_adjustment += 7;	\
+            break;				\
+          }					\
+          subnormal_prec_adjustment += 8;
 
 
 #define NEG_ZERO_BUG 196866 /* A bug affecting mpfr_fits_u*_p functions         */
@@ -263,3 +374,132 @@ typedef __float128 float128;
 #define mp_bitcnt_t unsigned long int
 #endif
 
+/* For nvtoa() */
+#if defined(NV_IS_53_BIT)
+#define MATH_MPFR_MAX_DIG 17
+#define MATH_MPFR_BITS 53
+#define MATH_MPFR_NV_MAX 1.7976931348623157e+308
+#define MATH_MPFR_NORMAL_MIN 2.2250738585072014e-308
+
+# if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
+
+# define D_CONDITION_1 i<=7
+# define D_INC_OR_DEC i++;
+# define DIND_0 0
+# define DIND_1 1
+
+# else						/* little endian architecture */
+
+# define D_CONDITION_1 i>=0
+# define D_INC_OR_DEC i--;
+# define DIND_0 7
+# define DIND_1 6
+
+# endif
+
+#elif defined(NV_IS_LONG_DOUBLE) && REQUIRED_LDBL_MANT_DIG == 64
+#define MATH_MPFR_MAX_DIG 21
+#define MATH_MPFR_BITS 64
+#define MATH_MPFR_NV_MAX 1.18973149535723176502e4932L
+#define MATH_MPFR_NORMAL_MIN 3.36210314311209350626e-4932L
+
+# if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
+
+# define LD_CONDITION_1 i<=9
+# define LD_INC_OR_DEC i++;
+# define LDIND_0 0
+# define LDIND_1 1
+# define LDIND_2 2
+
+# else						/* little endian architecture */
+
+# define LD_CONDITION_1 i>=0
+# define LD_INC_OR_DEC i--;
+# define LDIND_0 9
+# define LDIND_1 8
+# define LDIND_2 7
+
+# endif
+
+#elif defined(NV_IS_LONG_DOUBLE) && REQUIRED_LDBL_MANT_DIG == 2098
+#define MATH_MPFR_MAX_DIG 33
+#define MATH_MPFR_BITS 2098
+#define MATH_MPFR_NV_MAX 1.797693134862315807937289714053e+308L
+#define MATH_MPFR_NORMAL_MIN 2.2250738585072014e-308
+
+# if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
+# define DD_CONDITION_1 i<=7
+# define DD_CONDITION_2 i=2;i<8;i++
+# define DD_INC_OR_DEC i++;
+# define IND_0 0
+# define IND_1 1
+# define LSD_BYTE_1 8
+# define LSD_BYTE_2 9
+# define LSD_BYTE_3 10
+# define LSD_BYTE_4 11
+# define LSD_BYTE_5 12
+# define LSD_BYTE_6 13
+# define LSD_BYTE_7 14
+# define LSD_BYTE_8 15
+
+# else                                        /* little endian architecture */
+
+# define DD_CONDITION_1 i>=0
+# define DD_CONDITION_2 i=13;i>7;i--
+# define DD_INC_OR_DEC i--;
+# define IND_0 15
+# define IND_1 14
+# define LSD_BYTE_1 7
+# define LSD_BYTE_2 6
+# define LSD_BYTE_3 5
+# define LSD_BYTE_4 4
+# define LSD_BYTE_5 3
+# define LSD_BYTE_6 2
+# define LSD_BYTE_7 1
+# define LSD_BYTE_8 0
+
+# endif /* MPFR_HAVE_BENDIAN */
+
+#else
+#define MATH_MPFR_MAX_DIG 36
+#define MATH_MPFR_BITS 113
+#define MATH_MPFR_NV_MAX 1.18973149535723176508575932662800702e+4932Q
+#define MATH_MPFR_NORMAL_MIN 3.3621031431120935062626778173217526e-4932Q
+
+# if defined(MPFR_HAVE_BENDIAN)                /* big endian architecture - defined by Makefile.PL */
+
+# define Q_CONDITION_1 i<=15
+# define Q_INC_OR_DEC i++;
+# define QIND_0 0
+# define QIND_1 1
+# define QIND_2 2
+
+# else						/* little endian architecture */
+
+# define Q_CONDITION_1 i>=0
+# define Q_INC_OR_DEC i--;
+# define QIND_0 15
+# define QIND_1 14
+# define QIND_2 13
+
+# endif
+
+#endif
+
+/* End of defines for nvtoa() */
+
+#define NEW_MATH_MPFR_OBJECT(PACNAME,FUNCNAME) \
+     Newx(mpfr_t_obj, 1, mpfr_t);							\
+     if(mpfr_t_obj == NULL) croak("Failed to allocate memory in FUNCNAME function");	\
+     obj_ref = newSV(0);								\
+     obj = newSVrv(obj_ref, PACNAME);
+
+#define OBJ_READONLY_ON \
+     sv_setiv(obj, INT2PTR(IV,mpfr_t_obj));	\
+     SvREADONLY_on(obj);
+
+#define RETURN_STACK_2 \
+     ST(0) = sv_2mortal(obj_ref);	\
+     ST(1) = sv_2mortal(newSViv(ret));	\
+     XSRETURN(2);
+
